@@ -191,7 +191,12 @@ class KernelManager(ConnectionFileMixin):
             # but it should be.
             cmd[0] = sys.executable
 
-        ns = dict(connection_file=self.connection_file,
+        # Make sure to use the realpath for the connection_file
+        # On windows, when running with the store python, the connection_file path
+        # is not usable by non python kernels because the path is being rerouted when
+        # inside of a store app.
+        # See this bug here: https://bugs.python.org/issue41196
+        ns = dict(connection_file=os.path.realpath(self.connection_file),
                   prefix=sys.prefix,
                  )
 
@@ -459,7 +464,6 @@ class KernelManager(ConnectionFileMixin):
         This is a private method, callers should use shutdown_kernel(now=True).
         """
         if self.has_kernel:
-
             # Signal the kernel to terminate (sends SIGKILL on Unix and calls
             # TerminateProcess() on Win32).
             try:
@@ -483,8 +487,6 @@ class KernelManager(ConnectionFileMixin):
             # Block until the kernel terminates.
             self.kernel.wait()
             self.kernel = None
-        else:
-            raise RuntimeError("Cannot kill kernel. No kernel is running!")
 
     def interrupt_kernel(self):
         """Interrupts the kernel by sending it a signal.
@@ -709,8 +711,6 @@ class AsyncKernelManager(KernelManager):
                 if self.kernel is not None:
                     self.kernel.wait()
             self.kernel = None
-        else:
-            raise RuntimeError("Cannot kill kernel. No kernel is running!")
 
     async def interrupt_kernel(self):
         """Interrupts the kernel by sending it a signal.
