@@ -17,39 +17,19 @@ import hashlib
 import hmac
 import logging
 import os
+import pickle
 import pprint
 import random
 import warnings
 from datetime import datetime
 
-try:
-    import cPickle
-    pickle = cPickle
-except:
-    cPickle = None
-    import pickle
+PICKLE_PROTOCOL = pickle.DEFAULT_PROTOCOL
 
-try:
-    # py3
-    PICKLE_PROTOCOL = pickle.DEFAULT_PROTOCOL
-except AttributeError:
-    PICKLE_PROTOCOL = pickle.HIGHEST_PROTOCOL
+# We are using compare_digest to limit the surface of timing attacks
+from hmac import compare_digest
 
-try:
-    # We are using compare_digest to limit the surface of timing attacks
-    from hmac import compare_digest
-except ImportError:
-    # Python < 2.7.7: When digests don't match no feedback is provided,
-    # limiting the surface of attack
-    def compare_digest(a,b): return a == b
-
-try:
-    from datetime import timezone
-    utc = timezone.utc
-except ImportError:
-    # Python 2
-    from dateutil.tz import tzutc
-    utc = tzutc()
+from datetime import timezone
+utc = timezone.utc
 
 import zmq
 from zmq.utils import jsonapi
@@ -59,11 +39,10 @@ from zmq.eventloop.zmqstream import ZMQStream
 from traitlets.config.configurable import Configurable, LoggingConfigurable
 from ipython_genutils.importstring import import_item
 from jupyter_client.jsonutil import extract_dates, squash_dates, date_default
-from ipython_genutils.py3compat import (str_to_bytes, str_to_unicode, unicode_type,
-                                     iteritems)
-from traitlets import (CBytes, Unicode, Bool, Any, Instance, Set,
-                                        DottedObjectName, CUnicode, Dict, Integer,
-                                        TraitError,
+from ipython_genutils.py3compat import str_to_bytes, str_to_unicode
+from traitlets import (
+    CBytes, Unicode, Bool, Any, Instance, Set, DottedObjectName, CUnicode,
+    Dict, Integer, TraitError, observe
 )
 from jupyter_client import protocol_version
 from jupyter_client.adapter import adapt
@@ -79,12 +58,12 @@ def squash_unicode(obj):
     if isinstance(obj,dict):
         for key in obj.keys():
             obj[key] = squash_unicode(obj[key])
-            if isinstance(key, unicode_type):
+            if isinstance(key, str):
                 obj[squash_unicode(key)] = obj.pop(key)
     elif isinstance(obj, list):
         for i,v in enumerate(obj):
             obj[i] = squash_unicode(v)
-    elif isinstance(obj, unicode_type):
+    elif isinstance(obj, str):
         obj = obj.encode('utf8')
     return obj
 
@@ -129,7 +108,7 @@ def new_id():
     id string (16 random bytes as hex-encoded text, chunks separated by '-')
     """
     buf = os.urandom(16)
-    return u'-'.join(b2a_hex(x).decode('ascii') for x in (
+    return '-'.join(b2a_hex(x).decode('ascii') for x in (
         buf[:4], buf[4:]
     ))
 
@@ -180,23 +159,25 @@ class SessionFactory(LoggingConfigurable):
     """
 
     logname = Unicode('')
-    def _logname_changed(self, name, old, new):
-        self.log = logging.getLogger(new)
+
+    @observe('logname')
+    def _logname_changed(self, change):
+        self.log = logging.getLogger(change['new'])
 
     # not configurable:
     context = Instance('zmq.Context')
     def _context_default(self):
-        return zmq.Context.instance()
+        return zmq.Context()
 
     session = Instance('jupyter_client.session.Session',
                        allow_none=True)
 
-    loop = Instance('zmq.eventloop.ioloop.IOLoop')
+    loop = Instance('tornado.ioloop.IOLoop')
     def _loop_default(self):
-        return IOLoop.instance()
+        return IOLoop.current()
 
     def __init__(self, **kwargs):
-        super(SessionFactory, self).__init__(**kwargs)
+        super().__init__(**kwargs)
 
         if self.session is None:
             # construct the session
@@ -211,14 +192,14 @@ class Message(object):
 
     def __init__(self, msg_dict):
         dct = self.__dict__
-        for k, v in iteritems(dict(msg_dict)):
+        for k, v in dict(msg_dict).items():
             if isinstance(v, dict):
                 v = Message(v)
             dct[k] = v
 
     # Having this iterator lets dict(msg_obj) work out of the box.
     def __iter__(self):
-        return iter(iteritems(self.__dict__))
+        return self.__dict__.items()
 
     def __repr__(self):
         return repr(self.__dict__)
@@ -300,10 +281,10 @@ class Session(Configurable):
     """
 
     debug = Bool(False, config=True, help="""Debug output in the Session""")
-    
+
     check_pid = Bool(True, config=True,
         help="""Whether to check PID to protect against calls after fork.
-        
+
         This check can be disabled if fork-safety is handled elsewhere.
         """)
 
@@ -311,7 +292,10 @@ class Session(Configurable):
             help="""The name of the packer for serializing messages.
             Should be one of 'json', 'pickle', or an import name
             for a custom callable serializer.""")
-    def _packer_changed(self, name, old, new):
+
+    @observe('packer')
+    def _packer_changed(self, change):
+        new = change['new']
         if new.lower() == 'json':
             self.pack = json_packer
             self.unpack = json_unpacker
@@ -326,7 +310,10 @@ class Session(Configurable):
     unpacker = DottedObjectName('json', config=True,
         help="""The name of the unpacker for unserializing messages.
         Only used with custom functions for `packer`.""")
-    def _unpacker_changed(self, name, old, new):
+
+    @observe('unpacker')
+    def _unpacker_changed(self, change):
+        new = change['new']
         if new.lower() == 'json':
             self.pack = json_packer
             self.unpack = json_unpacker
@@ -338,14 +325,15 @@ class Session(Configurable):
         else:
             self.unpack = import_item(str(new))
 
-    session = CUnicode(u'', config=True,
+    session = CUnicode('', config=True,
         help="""The UUID identifying this session.""")
     def _session_default(self):
         u = new_id()
         self.bsession = u.encode('ascii')
         return u
 
-    def _session_changed(self, name, old, new):
+    @observe('session')
+    def _session_changed(self, change):
         self.bsession = self.session.encode('ascii')
 
     # bsession is the session as bytes
@@ -368,28 +356,33 @@ class Session(Configurable):
     def _key_default(self):
         return new_id_bytes()
 
-    def _key_changed(self):
+    @observe('key')
+    def _key_changed(self, change):
         self._new_auth()
 
     signature_scheme = Unicode('hmac-sha256', config=True,
         help="""The digest scheme used to construct the message signatures.
         Must have the form 'hmac-HASH'.""")
-    def _signature_scheme_changed(self, name, old, new):
+
+    @observe('signature_scheme')
+    def _signature_scheme_changed(self, change):
+        new = change['new']
         if not new.startswith('hmac-'):
             raise TraitError("signature_scheme must start with 'hmac-', got %r" % new)
         hash_name = new.split('-', 1)[1]
         try:
             self.digest_mod = getattr(hashlib, hash_name)
-        except AttributeError:
-            raise TraitError("hashlib has no such attribute: %s" % hash_name)
+        except AttributeError as e:
+            raise TraitError("hashlib has no such attribute: %s" %
+                             hash_name) from e
         self._new_auth()
 
     digest_mod = Any()
     def _digest_mod_default(self):
         return hashlib.sha256
-    
+
     auth = Instance(hmac.HMAC, allow_none=True)
-    
+
     def _new_auth(self):
         if self.key:
             self.auth = hmac.HMAC(self.key, digestmod=self.digest_mod)
@@ -406,8 +399,10 @@ class Session(Configurable):
 
     keyfile = Unicode('', config=True,
         help="""path to file containing execution key.""")
-    def _keyfile_changed(self, name, old, new):
-        with open(new, 'rb') as f:
+
+    @observe('keyfile')
+    def _keyfile_changed(self, change):
+        with open(change['new'], 'rb') as f:
             self.key = f.read().strip()
 
     # for protecting against sends from forks
@@ -416,13 +411,19 @@ class Session(Configurable):
     # serialization traits:
 
     pack = Any(default_packer) # the actual packer function
-    def _pack_changed(self, name, old, new):
+
+    @observe('pack')
+    def _pack_changed(self, change):
+        new = change['new']
         if not callable(new):
             raise TypeError("packer must be callable, not %s"%type(new))
 
     unpack = Any(default_unpacker) # the actual packer function
-    def _unpack_changed(self, name, old, new):
+
+    @observe('unpack')
+    def _unpack_changed(self, change):
         # unpacker is not checked - it is assumed to be
+        new = change['new']
         if not callable(new):
             raise TypeError("unpacker must be callable, not %s"%type(new))
 
@@ -478,7 +479,7 @@ class Session(Configurable):
             The file containing a key.  If this is set, `key` will be
             initialized to the contents of the file.
         """
-        super(Session, self).__init__(**kwargs)
+        super().__init__(**kwargs)
         self._check_packers()
         self.none = self.pack({})
         # ensure self._session_default() if necessary, so bsession is defined:
@@ -488,10 +489,30 @@ class Session(Configurable):
         if not self.key:
             get_logger().warning("Message signing is disabled.  This is insecure and not recommended!")
 
+    def clone(self):
+        """Create a copy of this Session
+
+        Useful when connecting multiple times to a given kernel.
+        This prevents a shared digest_history warning about duplicate digests
+        due to multiple connections to IOPub in the same process.
+
+        .. versionadded:: 5.1
+        """
+        # make a copy
+        new_session = type(self)()
+        for name in self.traits():
+            setattr(new_session, name, getattr(self, name))
+        # fork digest_history
+        new_session.digest_history = set()
+        new_session.digest_history.update(self.digest_history)
+        return new_session
+
+    message_count = 0
     @property
     def msg_id(self):
-        """always return new uuid"""
-        return new_id()
+        message_number = self.message_count
+        self.message_count += 1
+        return '{}_{}'.format(self.session, message_number)
 
     def _check_packers(self):
         """check packers for datetime support."""
@@ -510,7 +531,7 @@ class Session(Configurable):
                 jsonmsg = ""
             raise ValueError(
                 msg.format(packer=self.packer, e=e, jsonmsg=jsonmsg)
-            )
+            ) from e
 
         # ensure packed message is bytes
         if not isinstance(packed, bytes):
@@ -528,7 +549,7 @@ class Session(Configurable):
                 jsonmsg = ""
             raise ValueError(
                 msg.format(packer=self.packer, unpacker=self.unpacker, e=e, jsonmsg=jsonmsg)
-            )
+            ) from e
 
         # check datetime support
         msg = dict(t=utcnow())
@@ -608,7 +629,7 @@ class Session(Configurable):
         elif isinstance(content, bytes):
             # content is already packed, as in a relayed message
             pass
-        elif isinstance(content, unicode_type):
+        elif isinstance(content, str):
             # should be bytes, but JSON often spits out unicode
             content = content.encode('utf8')
         else:
@@ -699,13 +720,20 @@ class Session(Configurable):
             )
             return
         buffers = [] if buffers is None else buffers
-        for buf in buffers:
-            if not isinstance(buf, memoryview):
+        for idx, buf in enumerate(buffers):
+            if isinstance(buf, memoryview):
+                view = buf
+            else:
                 try:
                     # check to see if buf supports the buffer protocol.
-                    memoryview(buf)
-                except TypeError:
-                    raise TypeError("Buffer objects must support the buffer protocol.")
+                    view = memoryview(buf)
+                except TypeError as e:
+                    raise TypeError("Buffer objects must support the buffer protocol.") from e
+            # memoryview.contiguous is new in 3.3,
+            # just skip the check on Python 2
+            if hasattr(view, 'contiguous') and not view.contiguous:
+                # zmq requires memoryviews to be contiguous
+                raise ValueError("Buffer %i (%r) is not contiguous" % (idx, buf))
 
         if self.adapt_version:
             msg = adapt(msg, self.adapt_version)
@@ -754,7 +782,8 @@ class Session(Configurable):
             to_send.extend(ident)
 
         to_send.append(DELIM)
-        to_send.append(self.sign(msg_list))
+        # Don't include buffers in signature (per spec).
+        to_send.append(self.sign(msg_list[0:4]))
         to_send.extend(msg_list)
         stream.send_multipart(to_send, flags, copy=copy)
 
